#include "controller.h"

#include "gyro.h"
#include "hall.h"
#include "adc.h"
#include "actuator.h"
#include "matrix.h"

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#define DT 0.008f

volatile bool running = 1;
volatile bool ready = 1;

Matrix33 A_p = {0.01304,  8.474e-006, -5.219e-006,
                0.002481,  0.9923,  0.006574,
                0.4202,   -1.808,  0.6636};

Vector3 B_p = {0.2121, 
               0.0004911,
               0.1541};

Vector3 C_p = {0, 1, 0}; //Technically it is a row vector, but it doesn't really
						 //matter cause of the way it is used.

//Kalman gain, maybe it should not be a constant.
Vector3 K_p = {0.4174,
			   0.2084,
			   3.0020};

void CONTROLLER_init() {

	ADC_init();
	ACTUATOR_init();

	//Start timer to run the controller periodically:
	//CTC mode, generate a timer interrupt each time the timer reaches the
	//top value
	TCCR0 |= (1 << WGM21);

	//Period of 8 ms, set by prescaler and TOP value: 

	//Prescaler clk/1024:
	TCCR0 |= (1 << CS02) | (1 << CS01) | (1 << CS00);

	//TOP value of 64
	OCR0 = 0x3F; //TODO temp

}

void CONTROLLER_run() {

	Vector3 xEst_p = {0, 0, 0};
	Vector3 xEstMinus_p = {0, 0, 0};
	float u_p = 0;
	float integralE_p = 0;
	float y_p = 0;
	
	GYRO_calibrate(GYRO_P);

	//Enable compare match interrupt:
	TIMSK |= (1 << OCIE0);

	while(running) {
		while(!ready);

		//Measurements
		float sensor_p = HALL_getValue(HALL_P);
		float gyro_p = GYRO_getValue(GYRO_P);

		//fortegn?
		float z_p = sensor_p;

		//Kalman filter, calculate state estimates
		xEstMinus_p = vecplusvec( matxvec(A_p, xEst_p), vecxscalar(B_p, u_p) );
		y_p = z_p - innerproduct(C_p, xEstMinus_p);
		xEst_p = vecplusvec( xEstMinus_p, vecxscalar(K_p, y_p) );

		integralE_p += (gyro_p + xEst_p.v2) * DT;

		//Controller
		float P = 2;
		float I = 200;
		float D = 1;
		u_p = -P*(gyro_p + xEst_p.v2) -I*integralE_p - D*xEst_p.v3;

		u_p = ACTUATOR_setValue(ACTUATOR_P, u_p);

		//printf("%ld\t%ld\t%ld\t%ld\t", (int32_t)(1000*xEst_p.v1), (int32_t)(1000*xEst_p.v2), (int32_t)(1000*xEst_p.v3), (int32_t)(1000*integralE_p));
		//printf("%d\t%d\t%d\n", (int16_t)(sensor_p*1000), (int16_t)(gyro_p*1000), (int16_t)(u_p*1000));

		//Finished with one cycle, wait for the next
		ready = 0;

	}
}

ISR(TIMER0_COMP_vect) {
	if(ready == 1) {
		printf("Longer period required\n");
	}
	ready = 1;
}
